/* $Id: xbuf_descriptor.c,v 1.1 2001/10/12 17:18:16 linnj Exp $ */
/******************************************************************************
*
*                           ****************************
*                          ** Copyright Xilinx, Inc. **
*                           ** All rights reserved.   **
*                           ****************************
*
* FILENAME:
*
* xbuf_descriptor.c
*
* DESCRIPTION:
*
* This file contains all processing for the buffer descriptor component.
* The buffer descriptor component is a passive component that only maps over
* a buffer descriptor data structure shared by the scatter gather DMA hardware
* and software. The component's primary purpose is to provide encapsulation of
* the buffer descriptor processing.
*
* A buffer descriptor is used by a DMA channel in scatter gather mode.  It
* describes a buffer which is to be transferred to or from a device.
* The processing is not implemented as a structure, but the following structure
* illustrates an example of a buffer descriptor.
*
*	struct
*	{
*		Xuint32 Control;
*		Xuint32 SourceAddress;
*		Xuint32 DestinationAddress;
*		Xuint32 Length;
*		Xuint32 Status;
*		Xuint32 DeviceStatus
*		Xuint32 NextPtr;
*		Xuint32 Id;
*		Xuint32 Flags;
*		Xuint32 RequestedLength;
*	}
*
* The control, source address, destination address, and length fields are
* copied into the respective registers of the DMA channel by the DMA channel
* when a buffer descriptor is processed.  Upon completion of processing for
* the buffer descriptor, the DMA channel copies it's status register and the
* device specific register back into the buffer descriptor.  These fields are
* used for interrupt processing by the device driver for the device.
*
* The buffer descriptor is tightly coupled with the DMA channel component
* since many of the fields of the buffer descriptor are put into or retrieved
* from the registers of a DMA channel.
*
* Some of the fields of a buffer descriptor are for software use only, such as
* the ID, flags, and requested length fields. The ID field is provided to
* allow a device driver to correlate the buffer descriptor to other data
* structures which may be operating system specific, such as a pointer to
* a higher level memory block.  The flags field currently only contains the
* locked flag.
*
* A buffer descriptor may be locked when it is retrieved from a scatter list.
* The affect of the lock is described in the DMA channel component. The buffer
* descriptor is unlocked at a later time when the buffer which the buffer
* descriptor points to has been completely processed.
*
* The requested length field allows the actual length of the DMA operation
* to be determined after it is complete.  The length field of the buffer
* descriptor is updated by the DMA channel hardware to indicate the number of
* bytes left to transfer.  The requested length field contains the original
* requested length such that the length field may be subtracted from it to
* get the actual length transferred.
*
* The buffer descriptor component is utilized together with a DMA channel.
* The DMA channel determines the restrictions which apply to the buffers
* pointed to by a buffer descriptor, such as alignment and caching.
*
* NOTES:
*
* Most of the functions of this component are implemented as macros in order
* to optimize the processing.  Buffer descriptors are typically used heavily
* in an interrupt context such that execution time is critical.
*
******************************************************************************/


/***************************** Include Files *********************************/

#include "xbuf_descriptor.h"

/************************** Constant Definitions *****************************/


/**************************** Type Definitions *******************************/


/***************** Macros (Inline Functions) Definitions *********************/


/************************** Function Prototypes ******************************/


/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_Initialize
*
* DESCRIPTION:
*
* This function initializes a buffer descriptor component by zeroing all of the
* fields of the buffer descriptor.  This function should be called prior to
* using a buffer descriptor.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_GetControl
*
* DESCRIPTION:
*
* This function gets the control field of a buffer descriptor component.  The
* DMA channel hardware transfers the control field from the buffer descriptor
* into the DMA control register when a buffer descriptor is processed.  It
* controls the details of the DMA transfer.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* The control field contents of the buffer descriptor. One or more of the
* following values may be contained the field.  Each of the values are
* unique bit masks.
*
*	XDC_DMACR_SOURCE_INCR_MASK	Increment the source address
*	XDC_DMACR_DEST_INCR_MASK	Increment the destination address
*	XDC_DMACR_SOURCE_LOCAL_MASK Local source address
*	XDC_DMACR_DEST_LOCAL_MASK	Local destination address
*	XDC_DMACR_SG_ENABLE_MASK	Scatter gather enable
*	XDC_DMACR_GEN_BD_INTR_MASK	Individual buffer descriptor interrupt
*	XDC_DMACR_LAST_BD_MASK		Last buffer descriptor in a packet
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_SetControl
*
* DESCRIPTION:
*
* This function sets the control field of a buffer descriptor component.  The
* DMA channel hardware transfers the control field from the buffer descriptor
* into the DMA control register when a buffer descriptor is processed.  It
* controls the details of the DMA transfer such as
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* Control contains the value to be written to the control field of the buffer
* descriptor. One or more of the following values may be contained the field.
* Each of the values are unique bit masks such that they may be ORed together
* to enable multiple bits or inverted and ANDed to disable multiple bits.
*
*	XDC_DMACR_SOURCE_INCR_MASK	Increment the source address
*	XDC_DMACR_DEST_INCR_MASK	Increment the destination address
*	XDC_DMACR_SOURCE_LOCAL_MASK Local source address
*	XDC_DMACR_DEST_LOCAL_MASK	Local destination address
*	XDC_DMACR_SG_ENABLE_MASK	Scatter gather enable
*	XDC_DMACR_GEN_BD_INTR_MASK	Individual buffer descriptor interrupt
*	XDC_DMACR_LAST_BD_MASK		Last buffer descriptor in a packet
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_IsLastControl
*
* DESCRIPTION:
*
* This function determines if this buffer descriptor is marked as being the
* last in the control field.  A packet may be broken up across multiple
* buffer descriptors such that the last buffer descriptor is the end of the
* packet.  The DMA channel hardware copies the control field from the buffer
* descriptor to the control register of the DMA channel when the buffer
* descriptor is processed.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* XTRUE if the buffer descriptor is marked as last in the control field,
* otherwise, XFALSE.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_SetLast
*
* DESCRIPTION:
*
* This function marks the buffer descriptor as being last in the control
* field of the buffer descriptor.  A packet may be broken up across multiple
* buffer descriptors such that the last buffer descriptor is the end of the
* packet.  The DMA channel hardware copies the control field from the buffer
* descriptor to the control register of the DMA channel when the buffer
* descriptor is processed.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_GetSrcAddress
*
* DESCRIPTION:
*
* This function gets the source address field of the buffer descriptor.
* The source address indicates the address of memory which is the
* source of a DMA scatter gather operation.  The DMA channel hardware
* copies the source address from the buffer descriptor to the source
* address register of the DMA channel when the buffer descriptor is processed.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* The source address field of the buffer descriptor.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_SetSrcAddress
*
* DESCRIPTION:
*
* This function sets the source address field of the buffer descriptor.
* The source address indicates the address of memory which is the
* source of a DMA scatter gather operation.  The DMA channel hardware
* copies the source address from the buffer descriptor to the source
* address register of the DMA channel when the buffer descriptor is processed.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* SourceAddress contains the source address field for the buffer descriptor.
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_GetDestAddress
*
* DESCRIPTION:
*
* This function gets the destination address field of the buffer descriptor.
* The destination address indicates the address of memory which is the
* destination of a DMA scatter gather operation.  The DMA channel hardware
* copies the destination address from the buffer descriptor to the destination
* address register of the DMA channel when the buffer descriptor is processed.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* The destination address field of the buffer descriptor.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_SetDestAddress
*
* DESCRIPTION:
*
* This function sets the destination address field of the buffer descriptor.
* The destination address indicates the address of memory which is the
* destination of a DMA scatter gather operation.  The DMA channel hardware
* copies the destination address from the buffer descriptor to the destination
* address register of the DMA channel when the buffer descriptor is processed.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* DestinationAddress contains the destination address field for the buffer
* descriptor.
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_GetLength
*
* DESCRIPTION:
*
* This function gets the length of the data transfer if the buffer descriptor
* has been processed by the DMA channel hardware.  If the buffer descriptor
* has not been processed, the return value will be zero indicating that no data
* has been transferred yet.  This function uses both the length and requested
* length fields of the buffer descriptor to determine the number of bytes
* transferred by the DMA operation. The length field of the buffer descriptor
* contains the number of bytes remaining from the requested length.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* The number of bytes which have been transferred by a DMA operation on the
* buffer descriptor.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_SetLength
*
* DESCRIPTION:
*
* This function sets the length and the requested length fields of the buffer
* descriptor.  The length field indicates the number of bytes to transfer for
* the DMA operation and the requested length is written with the same value.
* The requested length is not modified by the DMA hardware while the length
* field is modified by the hardware to indicate the number of bytes remaining
* in the transfer after the transfer is complete.  The requested length allows
* the software to calculate the actual number of bytes transferred for the DMA
* operation.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* Length contains the length to put in the length and requested length fields
* of the buffer descriptor.
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_GetStatus
*
* DESCRIPTION:
*
* This function gets the status field of a buffer descriptor component. The
* status field is written to the buffer descriptor by the DMA channel hardware
* after processing of a buffer descriptor is complete.  The status field
* indicates the status of the DMA operation.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* The status field contents of the buffer descriptor. One or more of the
* following values may be contained the field. Each of the values are
* unique bit masks.
*
*	XDC_DMASR_BUSY_MASK			The DMA channel is busy
*	XDC_DMASR_BUS_ERROR_MASK	A bus error occurred
*	XDC_DMASR_BUS_TIMEOUT_MASK	A bus timeout occurred
*	XDC_DMASR_LAST_BD_MASK		The last buffer descriptor of a packet
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_SetStatus
*
* DESCRIPTION:
*
* This function sets the status field of a buffer descriptor component.  The
* status field is written to the buffer descriptor by the DMA channel hardware
* after processing of a buffer descriptor is complete.  This function would
* typically be used during debugging of buffer descriptor processing.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* Status contains the status field for the buffer descriptor.
* The status register contents of the DMA channel. One or more of the
* following values may be contained the register. Each of the values are
* unique bit masks.
*
*	XDC_DMASR_BUSY_MASK			The DMA channel is busy
*	XDC_DMASR_BUS_ERROR_MASK	A bus error occurred
*	XDC_DMASR_BUS_TIMEOUT_MASK	A bus timeout occurred
*	XDC_DMASR_LAST_BD_MASK		The last buffer descriptor of a packet
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_IsLastStatus
*
* DESCRIPTION:
*
* This function determines if this buffer descriptor is marked as being the
* last in the status field.  A packet may be broken up across multiple
* buffer descriptors such that the last buffer descriptor is the end of the
* packet.  The DMA channel hardware copies the status register contents to
* the buffer descriptor of the DMA channel after processing of the buffer
* descriptor is complete.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* XTRUE if the buffer descriptor is marked as last in the status field,
* otherwise, XFALSE.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_GetDeviceStatus
*
* DESCRIPTION:
*
* This function gets the device status field of the buffer descriptor.  The
* device status is device specific such that the definition of the contents
* of this field are not defined in this function. The device is defined as the
* device which is using the DMA channel, such as an ethernet controller.  The
* DMA channel hardware copies the contents of the device status register into
* the buffer descriptor when processing of the buffer descriptor is complete.
* This value is typically used by the device driver for the device to determine
* the status of the DMA operation with respect to the device.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* The device status field of the buffer descriptor.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_SetDeviceStatus
*
* DESCRIPTION:
*
* This function sets the device status field of the buffer descriptor.  The
* device status is device specific such that the definition of the contents
* of this field are not defined in this function. The device is defined as the
* device which is using the DMA channel, such as an ethernet controller.  This
* function is typically only used for debugging/testing.
*
* The DMA channel hardware copies the contents of the device status register
* into the buffer descriptor when processing of the buffer descriptor is
* complete.  This value is typically used by the device driver for the device
* to determine the status of the DMA operation with respect to the device.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* Status contains the device status field for the buffer descriptor.
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_GetNextPtr
*
* DESCRIPTION:
*
* This function gets the next pointer field of the buffer descriptor.  This
* field is used to link the buffer descriptors together such that multiple DMA
* operations can be automated for scatter gather.  It also allows a single
* packet to be broken across multiple buffer descriptors.  The DMA channel
* hardware traverses the list of buffer descriptors using the next pointer
* of each buffer descriptor.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* The next pointer field of the buffer descriptor.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_SetNextPtr
*
* DESCRIPTION:
*
* This function sets the next pointer field of the buffer descriptor.  This
* field is used to link the buffer descriptors together such that many DMA
* operations can be automated for scatter gather.  It also allows a single
* packet to be broken across multiple buffer descriptors.  The DMA channel
* hardware traverses the list of buffer descriptors using the next pointer
* of each buffer descriptor.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* NextPtr contains the next pointer field for the buffer descriptor.
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_GetId
*
* DESCRIPTION:
*
* This function gets the ID field of the buffer descriptor.  The ID field is
* provided to allow a device driver to correlate the buffer descriptor to other
* data structures which may be operating system specific, such as a pointer to
* a higher level memory block. The ID field is not used by the DMA channel
* hardware and is application specific.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* The ID field of the buffer descriptor.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_SetId
*
* DESCRIPTION:
*
* This function sets the ID field of the buffer descriptor.  The ID field is
* provided to allow a device driver to correlate the buffer descriptor to other
* data structures which may be operating system specific, such as a pointer to
* a higher level memory block. The ID field is not used by the DMA channel
* hardware and is application specific.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* Id contains the ID field for the buffer descriptor.
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_GetFlags
*
* DESCRIPTION:
*
* This function gets the flags field of the buffer descriptor.  The flags
* field is not used by the DMA channel hardware and is used for software
* processing of buffer descriptors.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* The flags field of the buffer descriptor.  The field may contain one or more
* of the following values which are bit masks.
*
*	XBD_FLAGS_LOCKED_MASK		Indicates the buffer descriptor is locked
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_SetFlags
*
* DESCRIPTION:
*
* This function sets the flags field of the buffer descriptor.  The flags
* field is not used by the DMA channel hardware and is used for software
* processing of buffer descriptors.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* Flags contains the flags field for the buffer descriptor.  The field may
* contain one or more of the following values which are bit masks.
*
*	XBD_FLAGS_LOCKED_MASK		Indicates the buffer descriptor is locked
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_Lock
*
* DESCRIPTION:
*
* This function locks the buffer descriptor. A lock is specific to the
* scatter gather processing and prevents a buffer descriptor from being
* overwritten in the scatter gather list.  This field is not used by the DMA
* channel hardware such that the hardware could still write to the buffer
* descriptor.  Locking a buffer descriptor is application specific and not
* necessary to allow the DMA channel to use the buffer descriptor, but is
* provided for flexability in designing device drivers.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_Unlock
*
* DESCRIPTION:
*
* This function unlocks the buffer descriptor.  A lock is specific to the
* scatter gather processing and prevents a buffer descriptor from being
* overwritten in the scatter gather list.  This field is not used by the DMA
* channel hardware such that the hardware could still write to the buffer
* descriptor.  Locking a buffer descriptor is application specific and not
* necessary to allow the DMA channel to use the buffer descriptor, but is
* provided for flexability in designing device drivers.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* XBufDescriptor_IsLocked
*
* DESCRIPTION:
*
* This function determines if the buffer descriptor is locked.  The lock
* is not used by the DMA channel hardware and is used for software processing
* of buffer descriptors.
*
* ARGUMENTS:
*
* InstancePtr points to the buffer descriptor to operate on.
*
* RETURN VALUE:
*
* XTRUE if the buffer descriptor is locked, otherwise XFALSE.
*
* NOTES:
*
* Implemented as a macro in the include file.
*
******************************************************************************/
